@@ -10,6 +10,7 @@ import warnings
 from lucenequerybuilder import Q
 
 import options
+from query import QuerySequence, FilterSequence, CypherException
 from constants import (BREADTH_FIRST, DEPTH_FIRST,
                        STOP_AT_END_OF_GRAPH,
                        NODE_GLOBAL, NODE_PATH, NODE_RECENT,
@@ -76,11 +77,7 @@ class GraphDatabase(object):
             self.url = url
         else:
             self.url = "%s/" % url
-        response, content = None, None
-        try:
-            response, content = Request(**self._auth).get(url)
-        except Exception:
-            raise NotFoundError(result="Unable get root")
+        response, content = Request(**self._auth).get(self.url)
         if response.status == 200:
             response_json = json.loads(content)
             self._relationship_index = response_json['relationship_index']
@@ -89,11 +86,13 @@ class GraphDatabase(object):
             self._reference_node = response_json.get('reference_node', None)
             self._extensions_info = response_json['extensions_info']
             self._extensions = response_json['extensions']
+            self._cypher = response_json.get('cypher', None)
+            self.VERSION = response_json.get('neo4j_version', None)
             self.extensions = ExtensionsProxy(self._extensions,
                                               auth=self._auth)
             self.nodes = NodesProxy(self._node, self._reference_node,
                                     self._node_index,
-                                    auth=self._auth)
+                                    auth=self._auth, cypher=self._cypher)
             # Backward compatibility. The current style is more pythonic
             self.node = self.nodes
             # HACK: Neo4j doesn't provide the URLs to access to relationships
@@ -102,9 +101,13 @@ class GraphDatabase(object):
                                              url_parts[2])
             self.relationships = RelationshipsProxy(self._relationship,
                                                     self._relationship_index,
-                                                    auth=self._auth)
+                                                    auth=self._auth,
+                                                    cypher=self._cypher)
             self.Traversal = self._get_traversal_class()
-            self._batch = "%sbatch" % self.url
+            try:
+                self._batch = response_json["batch"]
+            except KeyError:
+                self._batch = "%sbatch" % self.url
         else:
             raise NotFoundError(response.status, "Unable get root")
 
@@ -180,6 +183,19 @@ class GraphDatabase(object):
             globals()[options.TX_NAME] = self._transactions[transaction_id]
         return self._transactions[transaction_id]
 
+    def query(self, q, params=None, returns=RAW):
+        if self._cypher:
+            types = {
+                "node": Node,
+                "relationship": Relationship,
+                "path": Path,
+                "position": Position,
+            }
+            return QuerySequence(self._cypher, self._auth, q=q, params=params,
+                                 types=types, returns=returns)
+        else:
+            raise CypherException
+
 
 class TransactionOperationProxy(dict, object):
     """
@@ -228,11 +244,13 @@ class TransactionOperationProxy(dict, object):
                               "This is needed in order to handle pickling in "
                               "nodes.",
                               DeprecationWarning)
+
                 def _create_relationship(*args, **kwargs):
                     _attr = "_create_relationship"
                     _func = object.__getattribute__(self, _attr)
                     _relationship = _func(attr)
                     return _relationship(*args, **kwargs)
+
                 return _create_relationship
         else:
             return object.__getattribute__(self, attr)
@@ -254,7 +272,7 @@ class TransactionOperationProxy(dict, object):
         return dict(self)
 
     def _set_properties(self, props={}):
-        _type = object.__getattribute__(self, "_extras")["type"]
+        # _type = object.__getattribute__(self, "_extras")["type"]
         if type == RELATIONSHIP:
             _body = dict.__getitem__(self, "body")
             _data = dict.__getitem__(_body, "data")
@@ -264,7 +282,7 @@ class TransactionOperationProxy(dict, object):
             _body.update(props)
 
     def _del_properties(self):
-        _type = object.__getattribute__(self, "_extras")["type"]
+        # _type = object.__getattribute__(self, "_extras")["type"]
         if type == RELATIONSHIP:
             _body = dict.__getitem__(self, "body")
             dict.__setitem__(_body, "data", {})
@@ -295,7 +313,7 @@ class TransactionOperationProxy(dict, object):
                 eltos = _proxy._list[key]
                 if _proxy._attribute:
                     return [_proxy._class(elto[_proxy._attribute],
-                                         update_dict=elto) for elto in eltos]
+                                          update_dict=elto) for elto in eltos]
                 else:
                     return [_proxy._class(elto) for elto in eltos]
             else:
@@ -304,11 +322,14 @@ class TransactionOperationProxy(dict, object):
             if _type == RELATIONSHIP:
                 _body = dict.__getitem__(self, "body")
                 return dict.__getitem__(_body, key)
+            elif _type == ITERABLE:
+                _body = dict.__getitem__(self, "body")
+                dict.__setitem__(self, "key", key)
+                return self
             else:
                 _body = dict.__getitem__(self, "body")
                 return dict.__getitem__(_body, key)
 
-
     def __setitem__(self, key, val):
         _type = object.__getattribute__(self, "_extras")["type"]
         _proxy = object.__getattribute__(self, "_proxy")
@@ -323,7 +344,6 @@ class TransactionOperationProxy(dict, object):
                 _body = dict.__getitem__(self, "body")
                 return dict.__setitem__(_body, key, val)
 
-
     def get_object(self):
         if self._object_ref:
             return self._object_ref()
@@ -350,16 +370,29 @@ class TransactionOperationProxy(dict, object):
             if not data["body"] or len(data["body"]) == 0:
                 self._proxy = list()
             else:
-                first_element = data["body"][0]
-                if "self" in first_element:
-                    if NODE in first_element["self"]:
-                        self._proxy = cls(Node, data["body"], "self",
-                                          auth=_auth)
-                    elif RELATIONSHIP in first_element["self"]:
-                        self._proxy = cls(Relationship, data["body"], "self",
-                                          auth=_auth)
+                self_keys = self().keys()
+                # Check if it is an element from the iterable
+                if "key" in self_keys and "of" in self_keys:
+                    _key = dict.__getitem__(self, "key")
+                    _of = dict.__getitem__(self, "of")
+                    _body = data["body"][_key]
+                    if _of == NODE:
+                        cls = Node
+                    else:  # Relationship
+                        cls = Relationship
+                    self._proxy = cls(_body["self"],
+                                      update_dict=_body, auth=_auth)
                 else:
-                    self._proxy = cls(Path, data["body"], auth=_auth)
+                    first_element = data["body"][0]
+                    if "self" in first_element:
+                        if NODE in first_element["self"]:
+                            self._proxy = cls(Node, data["body"], "self",
+                                              auth=_auth)
+                        elif RELATIONSHIP in first_element["self"]:
+                            self._proxy = cls(Relationship, data["body"],
+                                              "self", auth=_auth)
+                    else:
+                        self._proxy = cls(Path, data["body"], auth=_auth)
         else:
             if "self" in data["body"] and data["body"]["self"] != url:
                 self._proxy = cls(data["body"]["self"],
@@ -390,7 +423,8 @@ class TransactionOperationProxy(dict, object):
             object.__getattribute__(self, "_extras")["to"] = to
             create_relationship_url = "{%s}/relationships" % _job_id
             # Check if target node doesn't exist yet
-            if isinstance(to, TransactionOperationProxy):
+            if (isinstance(to, TransactionOperationProxy)
+                and not isinstance(to, Node)):
                 to_url = "{%s}" % to()["id"]
             else:
                 to_url = to.url
@@ -469,7 +503,6 @@ class TransactionOperationProxy(dict, object):
             return op
 
 
-
 class Transaction(object):
     """
     Transaction class.
@@ -580,7 +613,8 @@ class Transaction(object):
         else:
             return True
 
-    def subscribe(self, method, url, data=None, obj=None, returns=None):
+    def subscribe(self, method, url, data=None, obj=None, returns=None,
+                  of=None):
         job_id = len(self.operations)
         if url.startswith("{"):
             url_to = url
@@ -592,20 +626,23 @@ class Transaction(object):
             "method": method,
             "to": url_to,
             "id": job_id,
+            "of": of,
         }
         if data:
             params.update({"body": data})
         # Reunify PUT methods in just one
         transaction_operation = None
-        for i, operation in enumerate(self.operations):
-            if (operation()["method"] == params["method"] == TX_PUT
-                and operation()["to"] == params["to"]):
-                if "body" in operation():
-                    self.operations[i]()["body"].update(params["body"])
-                else:
-                    self.operations[i]()["body"] = params["body"]
-                transaction_operation = operation
-                break
+        if method == TX_PUT:
+            # TODO: Improve the performance of this search
+            for i, operation in enumerate(self.operations):
+                if (operation()["method"] == params["method"] == TX_PUT
+                    and operation()["to"] == params["to"]):
+                    if "body" in operation():
+                        self.operations[i]()["body"].update(params["body"])
+                    else:
+                        self.operations[i]()["body"] = params["body"]
+                    transaction_operation = operation
+                    break
         if not transaction_operation:
             transaction_operation = TransactionOperationProxy(obj=obj,
                                                               job=job_id,
@@ -657,12 +694,19 @@ class Base(object):
 
     def _update_dict_data(self):
         if "data" in self._dic:
-            self._dic["data"] = dict((self._safe_string(k),
-                                      self._safe_string(v))
+            self._dic["data"] = dict((Base._safe_string(k),
+                                      Base._safe_string(v))
                                       for k, v in self._dic["data"].items())
 
-    def _safe_string(self, s):
-        return unicode(s.decode("utf-8"))
+    @staticmethod
+    def _safe_string(s):
+        if isinstance(s, unicode):
+            return s
+        if isinstance(s, basestring):
+            return unicode(s.decode("utf-8"))
+        else:
+            # We avoid convert non-string values
+            return s
 
     def update(self, extensions=True, delete_on_not_found=False):
         if self._update_dict:
@@ -870,11 +914,13 @@ class NodesProxy(dict):
     create new nodes through calling.
     """
 
-    def __init__(self, node, reference_node=None, node_index=None, auth=None):
+    def __init__(self, node, reference_node=None, node_index=None, auth=None,
+                 cypher=None):
         self._node = node
         self._reference_node = reference_node
         self._node_index = node_index
         self._auth = auth or {}
+        self._cypher = cypher
 
     def __call__(self, **kwargs):
         tx = Transaction.get_transaction(kwargs.get("tx", None))
@@ -930,9 +976,14 @@ class NodesProxy(dict):
             node = self.__getitem__(key)
             del node
 
+    def filter(self, lookups=[], start=None):
+        return elements_filter(self, lookups=lookups, start=start,
+                               returns=Node)
+
     def _indexes(self):
         if self._node_index:
-            return IndexesProxy(self._node_index, NODE, auth=self._auth)
+            return IndexesProxy(self._node_index, NODE, auth=self._auth,
+                                cypher=self._cypher)
     indexes = property(_indexes)
 
 
@@ -1029,7 +1080,7 @@ class Node(Base):
             return self._dic["data"].viewitems()
         except AttributeError:
             return self._dic["data"].items()
-            
+
     def traverse(self, types=None, order=None, stop=None, returnable=None,
                  uniqueness=None, is_stop_node=None, is_returnable=None,
                  paginated=False, page_size=None, time_out=None,
@@ -1172,8 +1223,9 @@ class IndexesProxy(dict):
     Class proxy for indexes (nodes and relationships).
     """
 
-    def __init__(self, index_url, index_for=NODE, auth=None):
+    def __init__(self, index_url, index_for=NODE, auth=None, cypher=None):
         self._auth = auth or {}
+        self._cypher = cypher
         self.url = index_url
         self._index_for = index_for
         self._dict = self._get_dict()
@@ -1201,6 +1253,7 @@ class IndexesProxy(dict):
                     index_props[str(key)] = val
                 indexes_dict[index_name] = Index(self._index_for, index_name,
                                                  auth=self._auth,
+                                                 cypher=self._cypher,
                                                  **index_props)
             return indexes_dict
         elif response.status == 404:
@@ -1244,6 +1297,7 @@ class IndexesProxy(dict):
                         result_dict[str(key)] = val
                     self._dict[name] = Index(self._index_for, name,
                                              auth=self._auth,
+                                             cypher=self._cypher,
                                              **result_dict)
                 else:
                     msg = "Invalid data sent"
@@ -1278,6 +1332,104 @@ class IndexesProxy(dict):
         return self._dict.keys()
 
 
+class IndexKey(object):
+    """
+    Intermediate object so that lookups can be done like:
+    index[key][value]
+
+    Lookups are formated as http://.../{index_name}/{key}/{value}, so this
+    is the object that gets returned by index[key]. The REST request will
+    be sent when the value is specified.
+    """
+
+    def __init__(self, index_for, url, name, auth=None, cypher=None,
+                 key=None, tx=None):
+        self._auth = auth or {}
+        self._cypher = cypher
+        self._key = key
+        self._index_for = index_for
+        if url[-1] == '/':
+            url = url[:-1]
+        self.url = url
+        self.name = name
+        self.tx = tx
+
+    def __getitem__(self, value):
+        tx = None
+        if isinstance(value, Transaction):
+            tx = self.tx or value
+            value = value.get_value()
+        tx = Transaction.get_transaction(tx)
+        url = "%s/%s" % (self.url, smart_quote(value))
+        return Index._get_results(url, self._index_for, auth=self._auth,
+                                  tx=tx)
+
+    def __setitem__(self, value, item):
+        tx = self.tx
+        if isinstance(value, (list, tuple)):
+            tx = tx or value[1]
+            value = value[0]
+        if isinstance(item, Transaction):
+            tx = tx or item
+            item = item.get_value()
+        tx = Transaction.get_transaction(tx)
+        # Neo4j hardly crush if you try to index a relationship in a
+        # node index and viceversa.
+        is_node_index = (self._index_for == NODE
+                         and isinstance(item, Node))
+        is_relationship_index = (self._index_for == RELATIONSHIP
+                                 and isinstance(item, Relationship))
+        if not (is_node_index or is_relationship_index
+                or TransactionOperationProxy):
+            raise TypeError("%s is a %s and the index is for %ss"
+                            % (item, self._index_for.capitalize(),
+                               self._index_for))
+        if isinstance(item, Base):
+            url_ref = item.url
+        elif isinstance(item, TransactionOperationProxy):
+            url_ref = "{%s}" % item()["id"]
+        else:
+            url_ref = item
+        request_url_and_key = self.url.rsplit('/', 1)  # assumes a key
+        # It's URL encoded and we need Unicode
+        key = urllib.unquote(str(request_url_and_key[1])).decode("utf8")
+        data = {"key": key,
+                "value": value,  # smart_quote is not needed anymore
+                "uri": url_ref}
+        if tx:
+            request_url = "index/%s/%s" % (self._index_for, self.name)
+            op = tx.subscribe(TX_POST, request_url, data=data,
+                              obj=self, returns=self._index_for)
+            return op
+        else:
+            request = Request(**self._auth)
+            response, content = request.post(request_url_and_key[0],
+                                             data=data)
+            if response.status == 201:
+                # Returns object that was indexed
+                entity = json.loads(content)
+                if self._index_for == NODE:
+                    return Node(entity['self'], data=entity['data'],
+                                auth=self._auth, update_dict=entity)
+                else:
+                    return Relationship(entity['self'],
+                                        data=entity['data'],
+                                        auth=self._auth)
+            else:
+                raise StatusException(response.status,
+                                      "Error requesting index with POST " \
+                                      "%s, data %s" \
+                                      % (request_url_and_key[0], url_ref))
+
+    def query(self, value, tx=None):
+        url = "%s?query=%s" % (self.url, smart_quote(value))
+        return Index._get_results(url, self._index_for, auth=self._auth,
+                                  tx=tx)
+
+    def filter(self, lookups=[], value=None):
+        return Index._filter(self, lookups, self._key, value)
+
+
 class Index(object):
     """
     key/value indexed lookups. Create an index object with GraphDatabase.index.
@@ -1288,7 +1440,8 @@ class Index(object):
     def _get_results(url, node_or_rel, auth={}, tx=None):
         tx = Transaction.get_transaction(tx)
         if tx:
-            return tx.subscribe(TX_GET, url, obj=None, returns=ITERABLE)
+            return tx.subscribe(TX_GET, url, obj=None, returns=ITERABLE,
+                                of=node_or_rel)
         else:
             response, content = Request(**auth).get(url)
             if response.status == 200:
@@ -1305,99 +1458,9 @@ class Index(object):
                                         "Error requesting index with GET %s" \
                                         % url)
 
-    class IndexKey(object):
-        """
-        Intermediate object so that lookups can be done like:
-        index[key][value]
-
-        Lookups are formated as http://.../{index_name}/{key}/{value}, so this
-        is the object that gets returned by index[key]. The REST request will
-        be sent when the value is specified.
-        """
-
-        def __init__(self, index_for, url, name, auth=None, tx=None):
-            self._auth = auth or {}
-            self._index_for = index_for
-            if url[-1] == '/':
-                url = url[:-1]
-            self.url = url
-            self.name = name
-            self.tx = tx
-
-        def __getitem__(self, value):
-            tx = None
-            if isinstance(value, Transaction):
-                tx = self.tx or value
-                value = value.get_value()
-            tx = Transaction.get_transaction(tx)
-            url = "%s/%s" % (self.url, smart_quote(value))
-            return Index._get_results(url, self._index_for, auth=self._auth,
-                                      tx=tx)
-
-        def __setitem__(self, value, item):
-            tx = self.tx
-            if isinstance(value, (list, tuple)):
-                tx = tx or value[1]
-                value = value[0]
-            if isinstance(item, Transaction):
-                tx = tx or item
-                item = item.get_value()
-            tx = Transaction.get_transaction(tx)
-            # Neo4j hardly crush if you try to index a relationship in a
-            # node index and viceversa.
-            is_node_index = (self._index_for == NODE
-                             and isinstance(item, Node))
-            is_relationship_index = (self._index_for == RELATIONSHIP
-                                     and isinstance(item, Relationship))
-            if not (is_node_index or is_relationship_index
-                    or TransactionOperationProxy):
-                raise TypeError("%s is a %s and the index is for %ss"
-                                % (item, self._index_for.capitalize(),
-                                   self._index_for))
-            if isinstance(item, Base):
-                url_ref = item.url
-            elif isinstance(item, TransactionOperationProxy):
-                url_ref = "{%s}" % item()["id"]
-            else:
-                url_ref = item
-            request_url_and_key = self.url.rsplit('/', 1)  # assumes a key
-            # It's URL encoded and we need Unicode
-            key = urllib.unquote(str(request_url_and_key[1])).decode("utf8")
-            data = {"key": key,
-                    "value": value,  # smart_quote is not needed anymore
-                    "uri": url_ref}
-            if tx:
-                request_url = "index/%s/%s" % (self._index_for, self.name)
-                op = tx.subscribe(TX_POST, request_url, data=data,
-                                  obj=self, returns=self._index_for)
-                return op
-            else:
-                request = Request(**self._auth)
-                response, content = request.post(request_url_and_key[0],
-                                                 data=data)
-                if response.status == 201:
-                    # Returns object that was indexed
-                    entity = json.loads(content)
-                    if self._index_for == NODE:
-                        return Node(entity['self'], data=entity['data'],
-                                    auth=self._auth, update_dict=entity)
-                    else:
-                        return Relationship(entity['self'],
-                                            data=entity['data'],
-                                            auth=self._auth)
-                else:
-                    raise StatusException(response.status,
-                                          "Error requesting index with POST " \
-                                          "%s, data %s" \
-                                          % (request_url_and_key[0], url_ref))
-
-        def query(self, value, tx=None):
-            url = "%s?query=%s" % (self.url, smart_quote(value))
-            return Index._get_results(url, self._index_for, auth=self._auth,
-                                      tx=tx)
-
-    def __init__(self, index_for, name, auth=None, **kwargs):
+    def __init__(self, index_for, name, auth=None, cypher=None, **kwargs):
         self._auth = auth or {}
+        self._cypher = cypher
         self._index_for = index_for
         self.name = name
         self.template = kwargs.get("template")
@@ -1452,14 +1515,16 @@ class Index(object):
             tx = tx or key[1]
             key = key[0]
         tx = Transaction.get_transaction(tx)
-        key = smart_quote(key)
+        _key = smart_quote(key)
         if value:
             value = smart_quote(value)
-            return self.IndexKey(self._index_for, "%s/%s" % (self.url, key),
-                                name=self.name, auth=self._auth, tx=tx)[value]
+            return IndexKey(self._index_for, "%s/%s" % (self.url, _key),
+                            name=self.name, auth=self._auth,
+                            cypher=self._cypher, key=key, tx=tx)[value]
         else:
-            return self.IndexKey(self._index_for, "%s/%s" % (self.url, key),
-                                 name=self.name, auth=self._auth, tx=tx)
+            return IndexKey(self._index_for, "%s/%s" % (self.url, _key),
+                            name=self.name, auth=self._auth,
+                            cypher=self._cypher, key=key, tx=tx)
 
     def delete(self, key=None, value=None, item=None, tx=None):
         if not key and not value and not item:
@@ -1532,6 +1597,47 @@ class Index(object):
                                      'not include a field.')
                 return index_key.query(unicode(query))
 
+    def filter(self, lookups=[], key=None, value=None):
+        return Index._filter(self, lookups, key, value)
+
+    @staticmethod
+    def _filter(cls, lookups=[], key=None, value=None):
+        if cls._cypher:
+            if cls._index_for == NODE:
+                start = u"node:"
+                returns = Node
+            elif cls._index_for == RELATIONSHIP:
+                start = u"relationship:"
+                returns = Relationship
+            else:
+                raise CypherException("Index not valid")
+            index_name = Base._safe_string(cls.name)
+            start = u"%s`%s`" % (start, index_name.replace("`", "\\`"))
+            if key:
+                key = Base._safe_string(key)
+                if value:
+                    value = Base._safe_string(value)
+                    start = u"%s(\"%s:%s\")" % (start, key, value)
+                else:
+                    start = u"%s(\"%s:*\")" % (start, key)
+            elif value:
+                raise CypherException("Index key not valid")
+            else:
+                start = u"%s(\"*:*\")" % start
+            if not isinstance(lookups, (list, tuple)):
+                lookups = [lookups]
+            types = {
+                "node": Node,
+                "relationship": Relationship,
+                "path": Path,
+                "position": Position,
+            }
+            return FilterSequence(cls._cypher, cls._auth, start=start,
+                                  types=types, lookups=lookups,
+                                  returns=returns)
+        else:
+            raise CypherException
+
 
 class RelationshipsProxy(dict):
     """
@@ -1539,8 +1645,10 @@ class RelationshipsProxy(dict):
     and create new relationships through calling.
     """
 
-    def __init__(self, relationship, relationship_index, auth=None):
+    def __init__(self, relationship, relationship_index, auth=None,
+                 cypher=None):
         self._auth = auth or {}
+        self._cypher = cypher
         self._relationship = relationship
         self._relationship_index = relationship_index
 
@@ -1576,10 +1684,14 @@ class RelationshipsProxy(dict):
             relationship = self.__getitem__(key)
             del relationship
 
+    def filter(self, lookups=[], start=None):
+        return elements_filter(self, lookups=lookups, start=start,
+                               returns=Relationship)
+
     def _indexes(self):
         if self._relationship_index:
             return IndexesProxy(self._relationship_index, RELATIONSHIP,
-                                auth=self._auth)
+                                auth=self._auth, cypher=self._cypher)
     indexes = property(_indexes)
 
 
@@ -1596,7 +1708,7 @@ class Relationships(object):
         self._len = 0
 
     def __getattr__(self, relationship_type):
-        auth = object.__getattribute__(self, "_auth")
+        # auth = object.__getattribute__(self, "_auth")
 
         def get_relationships(types=None, *args, **kwargs):
             tx = Transaction.get_transaction(kwargs.get("tx", None))
@@ -1800,18 +1912,47 @@ class BaseInAndOut(object):
             'type': property(lambda self: attr),
         })()
 
-
 All = BaseInAndOut(direction=RELATIONSHIPS_ALL)
 Incoming = BaseInAndOut(direction=RELATIONSHIPS_IN)
 Outgoing = BaseInAndOut(direction=RELATIONSHIPS_OUT)
 Undirected = BaseInAndOut(direction="both")  # Deprecated, use "All" instead
 
+
 class Direction(object):
     ANY = All
     INCOMING = Incoming
     OUTGOING = Outgoing
 
 
+class ExtensionModule(dict):
+    def __init__(self, klass_name, auth):
+        self.klass_name = klass_name
+        self.auth = auth
+        self.cache = {}
+
+    def __repr__(self):
+        return self.__unicode__()
+
+    def __str__(self):
+        return self.__unicode__()
+
+    def __unicode__(self):
+        return u"<Neo4j %s: %s>" % (self.__class__.__name__, unicode(self.klass_name.keys()))
+
+    def __getitem__(self, attr):
+        return self.__getattr__(attr)
+
+    def get(self, attr):
+        return self.__getattr__(attr)
+
+    def __getattr__(self, attr):
+        if attr in self.cache:
+            return self.cache[attr]
+        else:
+            self.cache[attr] = Extension(self.klass_name[attr], auth=self.auth)
+            return self.cache[attr]
+
+
 class ExtensionsProxy(dict):
     """
     Class proxy for extensions in order to allow get an extension by module
@@ -1829,20 +1970,8 @@ class ExtensionsProxy(dict):
     def __getattr__(self, attr):
         if attr in self._dict:
             return self._dict[attr]
-        class_name = self._extensions[attr]
-        auth = self._auth
-        # Using an anonymous class
-        return type("ExtensionModule", (dict, ), {
-            '__str__': lambda self: self.__unicode__(),
-            '__repr__': lambda self: self.__unicode__(),
-            '__unicode__': lambda self: u"<Neo4j %s: %s>" \
-                                        % (self.__class__.__name__,
-                                           unicode(class_name.keys())),
-            '__getitem__': lambda self, _attr: self.__getattr__(_attr),
-            '__getattr__': lambda self, _attr: Extension(class_name[_attr],
-                                                         auth=auth),
-            'get': lambda self, _attr: self.__getattr__(_attr),
-        })()
+        self._dict[attr] = ExtensionModule(self._extensions[attr], self._auth)
+        return self._dict[attr]
 
     def __repr__(self):
         return self.__unicode__()
@@ -2005,6 +2134,41 @@ class Extension(object):
         return params_kwargs
 
 
+def elements_filter(cls, lookups=[], start=None, returns=None):
+    if isinstance(start, (Index, IndexKey)):
+        return start.filter(lookups=lookups)
+    elif cls._cypher:
+        if start:
+            starts = []
+            if not isinstance(start, (list, tuple)):
+                start = [start]
+            for start_element in start:
+                try:
+                    starts.append(unicode(start_element.id))
+                except AttributeError:
+                    starts.append(unicode(start_element))
+        else:
+            starts = u"*"
+        if returns is Node:
+            start = u"node(%s)" % u", ".join(starts)
+        elif returns is Relationship:
+            start = u"rel(%s)" % u", ".join(starts)
+        else:
+            raise CypherException
+        if not isinstance(lookups, (list, tuple)):
+            lookups = [lookups]
+        types = {
+            "node": Node,
+            "relationship": Relationship,
+            "path": Path,
+            "position": Position,
+        }
+        return FilterSequence(cls._cypher, cls._auth, start=start,
+                              types=types, lookups=lookups, returns=returns)
+    else:
+        raise CypherException
+
+
 def smart_quote(val):
     if isinstance(val, (bool, int, float, long)):
         return urllib.quote(json.dumps(val), safe="")
