@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 __author__ = "Javier de la Rosa"
 __license__ = "GPL 3"
-__version__ = "1.7.0"
+__version__ = "1.8.0"
 __email__ = "versae@gmail.com"
 __url__ = "https://github.com/versae/neo4j-rest-client"
 __description__ = """Object-oriented Python library to interact with """ \
@@ -42,6 +42,9 @@ INDEX_NODE = "index_node"
 INDEX_RELATIONSHIP = "index_relationship"
 INDEX_EXACT = "exact"
 INDEX_FULLTEXT = "fulltext"
+# Cypher ordering
+ASC = "asc"
+DESC = "desc"
 # Transactions
 TX_GET = "GET"
 TX_PUT = "PUT"
