@@ -4,23 +4,23 @@ try:
     import cPickle as pickle
 except:
     import pickle
-
-import client
-import constants
-import request
+import sys
 import unittest
+import os
 
+import constants
+import client
+import options
+import query
+import request
+
+
+NEO4J_URL = os.environ.get('NEO4J_URL', "http://localhost:7474/db/data/")
 
 class NodesTestCase(unittest.TestCase):
-
     def setUp(self):
-        self.url = "http://localhost:7474/db/data/"
+        self.url = NEO4J_URL
         self.gdb = client.GraphDatabase(self.url)
-        # self.secured_url = "http://<instance>.hosted.neo4j.org:7000/db/data/"
-        # self.username = "username"
-        # self.password = "password"
-        # self.gdb = client.GraphDatabase(self.secured_url, username=self.username,
-        #                                 password=self.password)
 
     def tearDown(self):
         import options as clientCacheDebug
@@ -125,6 +125,12 @@ class NodesTestCase(unittest.TestCase):
         n2 = self.gdb.node[n1.id]
         self.assertEqual(n1.properties, n2.properties)
 
+    def test_set_node_properties_numbers_set(self):
+        n1 = self.gdb.node()
+        properties = {"name": "Jimmy Doe", "age": 30, "active": False}
+        n1.properties = properties
+        self.assertEqual(n1.properties, properties)
+
     def test_set_node_property_safe(self):
         n1 = self.gdb.node(language="Español antigüillo")
         n1.set("Idioma de los subtítulos", "Español antigüillo")
@@ -199,6 +205,14 @@ class RelationshipsTestCase(NodesTestCase):
         rel = n1.Knows(n2, since=1970)
         self.assertEqual(rel.properties, {"since": 1970})
 
+    def test_set_relationships_properties_numbers_set(self):
+        n1 = self.gdb.node()
+        n2 = self.gdb.node()
+        rel = n1.Knows(n2)
+        properties = {"name": "Jimmy Doe", "age": 30, "active": False}
+        rel.properties = properties
+        self.assertEqual(rel.properties, properties)
+
     def test_create_relationship_functional_properties(self):
         n1 = self.gdb.nodes.create()
         n2 = self.gdb.nodes.create()
@@ -553,7 +567,7 @@ class TraversalsTestCase(IndexesTestCase):
         traverser = db.traversal()\
             .relationships('related_to')\
             .traverse(start_node)
-        # The graph is traversed as 
+        # The graph is traversed as
         # you loop through the result.
         for node in traverser.nodes:
             pass
@@ -682,14 +696,17 @@ class ExtensionsTestCase(TraversalsTestCase):
     def test_gremlin_extension_reference_node(self):
         # Assuming the GremlinPlugin installed
         ext = self.gdb.extensions.GremlinPlugin
-        n = ext.execute_script(script='g.v(0)')
-        self.assertTrue(isinstance(n, client.Node))
+        n = self.gdb.nodes.create()
+        gremlin_n = ext.execute_script(script='g.v(%s)' % n.id)
+        self.assertEqual(gremlin_n, n)
 
     def test_gremlin_extension_reference_node_returns(self):
         # Assuming the GremlinPlugin installed
         ext = self.gdb.extensions.GremlinPlugin
-        n = ext.execute_script(script='g.v(0)', returns=constants.NODE)
-        self.assertTrue(isinstance(n, client.Node))
+        n = self.gdb.nodes.create()
+        gremlin_n = ext.execute_script(script='g.v(%s)' % n.id,
+                                       returns=constants.NODE)
+        self.assertEqual(gremlin_n, n)
 
     def test_gremlin_extension_relationships(self):
         # Assuming the GremlinPlugin installed
@@ -724,9 +741,18 @@ class ExtensionsTestCase(TraversalsTestCase):
     def test_gremlin_extension_reference_raw_returns(self):
         # Assuming the GremlinPlugin installed
         ext = self.gdb.extensions.GremlinPlugin
-        n = ext.execute_script(script='g.v(0)', returns=constants.RAW)
-        self.assertTrue(isinstance(n, dict))
+        n = self.gdb.nodes.create(name="Test")
+        gremlin_n = ext.execute_script(script='g.v(%s)' % n.id,
+                                       returns=constants.RAW)
+        self.assertEqual(gremlin_n["data"], n.properties)
+        self.assertTrue(isinstance(gremlin_n, dict))
 
+    def test_gremlin_results_raw(self):
+        # Assuming the GremlinPlugin installed
+        ext = self.gdb.extensions.GremlinPlugin
+        n = ext.execute_script(script='results = [1,2]', returns=constants.RAW)
+        self.assertTrue(isinstance(n, list))
+        self.assertEqual(n, [1, 2])
 
 class TransactionsTestCase(ExtensionsTestCase):
 
@@ -798,9 +824,10 @@ class TransactionsTestCase(ExtensionsTestCase):
         })
 
     def test_transaction_get(self):
-        n1 = self.gdb.nodes.get(1)
+        n = self.gdb.nodes.create()
+        n1 = self.gdb.nodes.get(n.id)
         with self.gdb.transaction():
-            n2 = self.gdb.nodes.get(1)
+            n2 = self.gdb.nodes.get(n.id)
         self.assertTrue(isinstance(n1, client.Node))
         self.assertTrue(isinstance(n2, client.Node))
         self.assertEqual(n1, n2)
@@ -918,9 +945,13 @@ class TransactionsTestCase(ExtensionsTestCase):
     def test_transaction_conections(self):
         import options as clientDebug
         clientDebug.DEBUG = True
+        id_list = []
+        for i in range(5):
+            n = self.gdb.nodes.create(number=i)
+            id_list.append(n.id)
         nodes = []
         with self.gdb.transaction(commit=False) as tx:
-            for i in range(5):
+            for i in id_list:
                 nodes.append(self.gdb.node[i])
         tx.commit()
         clientDebug.DEBUG = False
@@ -933,7 +964,7 @@ class TransactionsTestCase(ExtensionsTestCase):
             rel["when"] = "January"
         self.assertEqual(rel.properties, {"when": "January"})
 
-    def test_transaction_create_relationship_functional_mixed2(self):
+    def test_transaction_create_relationship_functional_mixed1(self):
         n1 = self.gdb.node()
         with self.gdb.transaction():
             n2 = self.gdb.node()
@@ -949,7 +980,7 @@ class TransactionsTestCase(ExtensionsTestCase):
             rel["when"] = "January"
         self.assertEqual(rel.properties, {"when": "January"})
 
-    def test_transaction_create_relationship_functional(self):
+    def test_transaction_create_relationship_functional2(self):
         with self.gdb.transaction():
             n1 = self.gdb.node()
             n2 = self.gdb.node()
@@ -973,7 +1004,7 @@ class TransactionsTestCase(ExtensionsTestCase):
     # https://github.com/mhluongo/neo4j-rest-client/blob/master/neo4jrestclient/tests.py
     def test_transaction_index_creation(self):
         """
-        Tests whether indexes are properly created during a transaction. 
+        Tests whether indexes are properly created during a transaction.
         Asserts the creation also behaves transactionally (ie, not until
         commit).
         """
@@ -1027,7 +1058,7 @@ class TransactionsTestCase(ExtensionsTestCase):
         tx = self.gdb.transaction()
         index_hits = index['test2']['test2']
         tx.commit()
-        self.assertTrue(n1 == index_hits[:][-1])
+        self.assertTrue(n1 == index_hits[-1])
 
     def test_transaction_remove_node_from_index(self):
         index = self.gdb.nodes.indexes.create('index3')
@@ -1042,9 +1073,9 @@ class TransactionsTestCase(ExtensionsTestCase):
 
     def test_transaction_query_index_for_new_node(self):
         #test nodes created in transaction
-        
+
         index = self.gdb.nodes.indexes.create('index4%s' \
-                                              % datetime.now().microsecond)
+                                              % datetime.now().strftime('%s%f'))
         tx = self.gdb.transaction(using_globals=False)
         n4 = self.gdb.nodes.create(tx=tx)
         index.add('test3','test3', n4, tx=tx)
@@ -1135,31 +1166,81 @@ class TransactionsTestCase(ExtensionsTestCase):
 #        #a non-transactional traversal will be 1 node short
 #        self.assertEqual(len(traversal), len(nodes))
 
-    def test_transaction_create_nodes_relationship(self):
+    def test_transaction_create_1000_nodes_relationship(self):
         TAG_DICT = {}
         EDGE_DICT = {}
-        id1 = 1001
-        id2 = 1002
-        TAG_DICT[id1] = self.gdb.node.create(tag="tag1")
-        TAG_DICT[id2] = self.gdb.node.create(tag="tag2")
-        if not (id1, id2) in EDGE_DICT:
-            with self.gdb.transaction():
-                edge = self.gdb.node(name='EDGE_%04d_%04d' % (id1, id2),
-                                     type='edge', tag1=id1, tag2=id2)
-                tag1 = TAG_DICT[id1]
-                tag2 = TAG_DICT[id2]
-                edge.relationships.create("EDGE_TAG", tag1)
-                edge.relationships.create("EDGE_TAG", tag2)
-                RUN.relationships.create("RUN_EDGE", edge)
-                EDGE_DICT[(id1,id2)] = edge
-        else:
-            edge = EDGE_DICT[(id1,id2)]
+        operations = 0
+        with self.gdb.transaction() as tx:
+            for i in xrange(1, 1000):
+                operations = len(tx.operations)
+                id1 = i
+                id2 = i + 1
+                TAG_DICT[id1] = self.gdb.node.create(tag="tag1")
+                TAG_DICT[id2] = self.gdb.node.create(tag="tag2")
+                RUN = self.gdb.node.create(name="RUN")
+                if not (id1, id2) in EDGE_DICT:
+                    edge = self.gdb.node(name='EDGE_%04d_%04d' % (id1, id2),
+                                         type='edge', tag1=id1, tag2=id2)
+                    tag1 = TAG_DICT[id1]
+                    tag2 = TAG_DICT[id2]
+                    edge.relationships.create("EDGE_TAG", tag1)
+                    edge.relationships.create("EDGE_TAG", tag2)
+                    RUN.relationships.create("RUN_EDGE", edge)
+                    EDGE_DICT[(id1,id2)] = edge
+                else:
+                    edge = EDGE_DICT[(id1,id2)]
+        print i, operations
+        self.assertTrue(operations >= i)
 
     def test_transaction_access_node(self):
         frame = self.gdb.node.create()
         with self.gdb.transaction():
             edge = self.gdb.node(name='EDGE')
-        frame.FRAME_EDGE(edge)
+        rel = frame.FRAME_EDGE(edge)
+        self.assertTrue(isinstance(rel, client.Relationship))
+
+    # Test from http://stackoverflow.com/questions/11407546/
+    def test_transaction_index_access_create_relationship(self):
+        s = self.gdb.node.create(id=1)
+        d = self.gdb.node.create(id=2)
+        nidx = self.gdb.nodes.indexes.create('nodelist')
+        nidx.add('nid',1, s)
+        nidx.add('nid',2, d)
+        nodelist = [(1, 2)]
+        with self.gdb.transaction():
+            for s_id, d_id in nodelist:
+                sn = nidx['nid'][s_id][-1]
+                dn = nidx['nid'][d_id][-1]
+#                rel = sn.Follows(dn)
+#        self.assertTrue(isinstance(rel, client.Relationship))
+        self.assertEqual(s, sn)
+        self.assertEqual(d, dn)
+
+    # Test from https://github.com/versae/neo4j-rest-client/issues/69
+    def test_transaction_split(self):
+
+        with self.gdb.transaction():
+            a = self.gdb.nodes.create(name='a')
+            b = self.gdb.nodes.create(name='b')
+
+        with self.gdb.transaction():
+            a.relationships.create("Test", b)
+            c = self.gdb.nodes.create(name='c')
+            b.relationships.create("Test", c)
+            c.relationships.create("Test", a)
+
+        a = self.gdb.nodes[a.id]
+        b = self.gdb.nodes[b.id]
+        c = self.gdb.nodes[c.id]
+
+        rel_ab = a.relationships.outgoing()[0]
+        assert(rel_ab.start == a and rel_ab.end == b)
+
+        rel_bc = b.relationships.outgoing()[0]
+        assert(rel_bc.start == b and rel_bc.end == c)
+
+        rel_ca = c.relationships.outgoing()[0]
+        assert(rel_ca.start == c and rel_ca.end == a)
 
 
 class PickleTestCase(TransactionsTestCase):
@@ -1177,9 +1258,234 @@ class PickleTestCase(TransactionsTestCase):
         self.assertEqual(r, pickle.loads(p))
 
 
-class Neo4jPythonClientTestCase(PickleTestCase):
+class QueryTestCase(PickleTestCase):
+
+    def test_query_raw(self):
+        n1 = self.gdb.nodes.create(name="John")
+        n2 = self.gdb.nodes.create(name="William")
+        n1.knows(n2, since=1982)
+        q = """start n=node(*) return n"""
+        result = self.gdb.query(q=q)
+        self.assertTrue(result is not None)
+
+    def test_query_raw_returns(self):
+        n1 = self.gdb.nodes.create(name="John")
+        n2 = self.gdb.nodes.create(name="William")
+        n1.knows(n2, since=1982)
+        q = """start n=node(*) return n limit 2"""
+        results = self.gdb.query(q=q, returns=[client.Node])
+        row1, row2 = results
+        m1, m2 = row1[0], row2[0]
+        self.assertTrue(isinstance(m1, client.Node))
+        self.assertTrue(isinstance(m2, client.Node))
+
+    def test_query_raw_returns_tuple(self):
+        n1 = self.gdb.nodes.create(name="John")
+        n2 = self.gdb.nodes.create(name="William")
+        rel_type = u"rel%s" % unicode(datetime.now().strftime('%s%f'))
+        r = n1.relationships.create(rel_type, n2, since=1982)
+        q = """start n=node(*) match n-[r:%s]-() """ \
+            """return n, n.name, r, r.since""" % rel_type
+        results = self.gdb.query(q, returns=(client.Node, unicode,
+                                             client.Relationship))
+        for node, name, rel, date in results:
+            self.assertTrue(node in (n1, n2))
+            self.assertTrue(name in (u"John", u"William"))
+            self.assertEqual(rel, r)
+            self.assertEqual(date, 1982)
+
+    def test_query_params_returns_tuple(self):
+        n1 = self.gdb.nodes.create(name="John")
+        n2 = self.gdb.nodes.create(name="William")
+        rel_type = u"rel%s" % unicode(datetime.now().strftime('%s%f'))
+        r = n1.relationships.create(rel_type, n2, since=1982)
+        q = """start n=node(*) match n-[r:`{rel}`]-() """ \
+            """return n, n.name, r, r.since"""
+        params = {
+            "rel": rel_type,
+        }
+        results = self.gdb.query(q, params=params,
+                                 returns=(client.Node, unicode,
+                                          client.Relationship))
+        for node, name, rel, date in results:
+            self.assertTrue(node in (n1, n2))
+            self.assertTrue(name in (u"John", u"William"))
+            self.assertEqual(rel, r)
+            self.assertEqual(date, 1982)
+
+
+class FilterTestCase(QueryTestCase):
+
+    def test_filter_nodes(self):
+        Q = query.Q
+        for i in range(5):
+            self.gdb.nodes.create(name="William %s" % i)
+        lookup = Q("name", istartswith="william")
+        williams = self.gdb.nodes.filter(lookup)
+        self.assertTrue(len(williams) >= 5)
+
+    def test_filter_nodes_complex_lookups(self):
+        Q = query.Q
+        for i in range(5):
+            self.gdb.nodes.create(name="James", surname="Smith %s" % i)
+        lookups = (
+            Q("name", exact="James") &
+            (Q("surname", startswith="Smith") &
+             ~Q("surname", endswith="1"))
+        )
+        williams = self.gdb.nodes.filter(lookups)
+        self.assertTrue(len(williams) >= 5)
+
+    def test_filter_nodes_slicing(self):
+        Q = query.Q
+        for i in range(5):
+            self.gdb.nodes.create(name="William %s" % i)
+        lookup = Q("name", istartswith="william")
+        williams = self.gdb.nodes.filter(lookup)[:4]
+        self.assertTrue(len(williams) == 4)
+
+    def test_filter_nodes_ordering(self):
+        Q = query.Q
+        for i in range(5):
+            self.gdb.nodes.create(name="William", code=i)
+        lookup = Q("code", gte=2)
+        williams = self.gdb.nodes.filter(lookup).order_by("code",
+                                                          constants.DESC)
+        self.assertTrue(williams[-1]["code"] > williams[0]["code"])
+
+    def test_filter_nodes_nullable(self):
+        Q = query.Q
+        for i in range(5):
+            self.gdb.nodes.create(name="William %s" % i)
+        lookup = Q("name", istartswith="william", nullable=False)
+        williams = self.gdb.nodes.filter(lookup)[:10]
+        self.assertTrue(len(williams) > 5)
+
+    def test_filter_nodes_start(self):
+        Q = query.Q
+        nodes = []
+        for i in range(5):
+            nodes.append(self.gdb.nodes.create(name="William %s" % i))
+        lookup = Q("name", istartswith="william")
+        williams = self.gdb.nodes.filter(lookup, start=nodes)
+        self.assertTrue(len(williams) == 5)
+
+    def test_filter_nodes_start_index(self):
+        Q = query.Q
+        t = unicode(datetime.now().strftime('%s%f'))
+        index_name = "filter_nodes_start_index_%s" % t
+        index = self.gdb.nodes.indexes.create(name=index_name)
+        for i in range(5):
+            n = self.gdb.nodes.create(name="William %s" % i)
+            index["name"]["William"] = n
+        lookup = Q("name", istartswith="william")
+        williams = self.gdb.nodes.filter(lookup, start=index)
+        self.assertTrue(len(williams) == 5)
+        williams = self.gdb.nodes.filter(lookup, start=index["name"])
+        self.assertTrue(len(williams) == 5)
+
+    def test_filter_index_for_nodes(self):
+        Q = query.Q
+        n1 = self.gdb.nodes.create(name="Lemmy", band="Motörhead")
+        index = self.gdb.nodes.indexes.create(name="music")
+        index["bandś"]["Motörhead"] = n1
+        lookup = Q("name", icontains='Lemmy')
+        self.assertTrue(n1 in index.filter(lookup))
+        self.assertTrue(n1 in index.filter(lookup, key="bandś"))
+        self.assertTrue(n1 in index.filter(lookup, key="bandś",
+                                           value="Motörhead"))
+        self.assertTrue(n1 in index["bandś"].filter(lookup))
+        self.assertTrue(n1 in index["bandś"].filter(lookup, value="Motörhead"))
+
+    def test_filter_index_for_relationships(self):
+        Q = query.Q
+        n1 = self.gdb.nodes.create(name="Jóhn Doe", place="Texaś")
+        n2 = self.gdb.nodes.create(name="Míchael Doe", place="Tíjuana")
+        r1 = self.gdb.relationships.create(n1, "Hateś", n2, since=1995)
+        index = self.gdb.relationships.indexes.create(name="bróthers")
+        index["feelińg"]["háte"] = r1
+        lookup = Q("since", lte=2000)
+        self.assertTrue(r1 in index.filter(lookup))
+        self.assertTrue(r1 in index.filter(lookup, key="feelińg"))
+        self.assertTrue(r1 in index.filter(lookup, key="feelińg",
+                                           value="háte"))
+        self.assertTrue(r1 in index["feelińg"].filter(lookup))
+        self.assertTrue(r1 in index["feelińg"].filter(lookup, value="háte"))
+
+    def test_filter_relationships_start(self):
+        Q = query.Q
+        rels = []
+        for i in range(5):
+            n1 = self.gdb.nodes.create(name="William %s" % i)
+            n2 = self.gdb.nodes.create(name="Rose %s" % i)
+            r = n1.loves(n2, since=(1990 + i))
+            rels.append(r)
+        lookup = Q("since", lte=2000)
+        old_loves = self.gdb.relationships.filter(lookup, start=rels)
+        self.assertTrue(len(old_loves) >= 5)
+
+    def test_filter_relationships_start_index(self):
+        Q = query.Q
+        t = unicode(datetime.now().strftime('%s%f'))
+        index_name = "filter_relationships_start_index_%s" % t
+        index = self.gdb.relationships.indexes.create(name=index_name)
+        for i in range(5):
+            n1 = self.gdb.nodes.create(name="William %s" % i)
+            n2 = self.gdb.nodes.create(name="Rose %s" % i)
+            r = n1.loves(n2, since=(1990 + i))
+            index["since"][1990] = r
+        lookup = Q("since", lte=2000)
+        old_loves = self.gdb.relationships.filter(lookup, start=index)
+        self.assertTrue(len(old_loves) == 5)
+        old_loves = self.gdb.relationships.filter(lookup, start=index["since"])
+        self.assertTrue(len(old_loves) == 5)
+
+
+class Neo4jPythonClientTestCase(FilterTestCase):
     pass
 
+
+class FakeCache(object):
+    def __init__(self, called):
+        self.called = called
+
+    def get(self, key):
+        self.called['get'] = True
+        return None
+
+    def set(self, key, value):
+        self.called['set'] = True
+
+    def delete(self, key):
+        pass
+
+
+class XtraCacheTestCase(unittest.TestCase):
+    def setUp(self):
+        self.cache_called = {}
+        options.CACHE = True
+        options.CACHE_STORE = FakeCache(self.cache_called)
+        # reload modules now cache set
+        del sys.modules['neo4jrestclient.request']
+        del sys.modules['neo4jrestclient.client']
+        import client
+        gdb = client.GraphDatabase(NEO4J_URL)
+        gdb
+
+    def test_custom_cache_used(self):
+        self.assertTrue(self.cache_called['get'])
+        self.assertTrue(self.cache_called['set'])
+
+    def tearDown(self):
+        # leave everything as we found it
+        options.CACHE = False
+        options.CACHE_STORE = '.cache'
+        del sys.modules['neo4jrestclient.request']
+        del sys.modules['neo4jrestclient.client']
+        import client
+        client
+
+
 if __name__ == '__main__':
     test_loader = unittest.TestLoader()
     suite = test_loader.loadTestsFromTestCase(Neo4jPythonClientTestCase)
